﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------
	
nUI.Class.Button = nUI.Class:CreatePrototype();
local prototype  = nUI.Class.Button.prototype;

-------------------------------------------------------------------------------
-- local cache

local ActionHasRange     = ActionHasRange;
local CreateFrame        = CreateFrame;
local GetBindingKey      = GetBindingKey;
local GetActionInfo      = GetActionInfo;
local IsActionInRange    = IsActionInRange;
local IsUsableAction     = IsUsableAction;
local IsConsumableAction = IsConsumableAction;
local GetActionCount     = GetActionCount;
local UnitExists         = UnitExists;
local UnitIsUnit         = UnitIsUnit;
local GetTime            = GetTime;
local GetActionInfo      = GetActionInfo;
local UIFrameFlash       = UIFrameFlash;
local floor              = math.floor;
local nUI                = nUI;
local SpellStatus        = nUI.SpellStatus;
local Cooldowns          = nUI.Cooldowns;

-------------------------------------------------------------------------------
-- spells we know to be valid for checking the global cooldown (GCD)

local nUI_GCDSpells =
{
};

local nUI_CandidateGCD = {};

function nUI_DurationText( remain )

	local text = "";
	
	if remain > 3600 then -- over an hour remaining
		
		local hours = remain / 3600;
		text = ("%0.0f"):format( hours ).."h";
		
	elseif remain > 60 then -- over one minute remaining
	
		local minutes = remain / 60;
		text = ("%0.0f"):format( minutes ).."m";
		
	elseif remain > 1 then -- more than a second to go
	
		local seconds = remain;
		text = ("%0.0f"):format( seconds ).."s";
		
	else -- less than two seconds
	
		text = ("%0.1f"):format( remain );
		
	end
	
	return text;
	
end

-------------------------------------------------------------------------------

function prototype:init( parent, id, layout )
	
	self.parent            = parent;
	self.id                = id;
	self.rid               = 1;
		
	self:layoutFrame( layout, true );
	
end

-------------------------------------------------------------------------------

function prototype:layoutFrame( layout, is_init )

	local	btn_name       = layout.button or layout.name.."Obj";
	
	if not is_init then self:unregisterEvents(); end

	self.elapsed           = 0;
	self.level             = self.parent:GetFrameLevel() + 1;
	self.layers            = {};	
	self.Timers            = {};
	self.Timers.range      = 0;
	self.Timers.cdc        = 0;
	self.Timers.cdc_expire = nil;
	self.Timers.visibility = 0;	
	self.Timers.scale      = 0;
	
	self.layout            = layout;
	self.width             = layout.width;
	self.height            = layout.height;
	self.actionType        = layout.btn_type;
	self.actionID          = layout.id or self.id;
	self.button            = _G[btn_name];
	self.is_usable         = true;
	self.is_oom            = false;
	self.is_oor            = false;
	self.is_cdc            = false;

	-- if we haven't created this object yet, then create the frames it needs
	
	if is_init then
		
		self.grid  = CreateFrame( "Frame", self.layout.name.."_Grid", self.parent );
		self.frame = CreateFrame( "Frame", self.layout.name, self.parent );
		
		-- the the button we're referencing does not exist, it's a custom button
		-- (we presume) so we create one for it
		
		if not self.button then
	
			local button      = CreateFrame( "CheckButton", btn_name, self.parent, self.layout.inherits );
	
			self.button       = button;
			button.buttonType = self.layout.btn_type;
			button.action     = self.id;
			button.showgrid   = 0;
			button.flashing   = 0;
			button.flashtime  = 0;
			custom_btn        = true;
	
			-- initialize the button
			
			local initBtn = function( button )
				this = button;
				ActionButton_OnLoad();
			end
			
			initBtn( button );
	
			button:SetAttribute( "action", self.id );
				
		end		

		self.layers.cdcFrame = CreateFrame( "Frame", self.layout.name.."_CDC", self.button );
		self.layers.cdc      = self.layers.cdcFrame:CreateFontString( "$parentLabel" );
	end
	
	-- the grid square that appears where the button should be when there is none
		
	self.grid:SetFrameLevel( 1 );
	self.grid:SetWidth( self.width );
	self.grid:SetHeight( self.height );
	self.grid:SetBackdrop(
		{
			bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
			edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
			tile     = true, 
			tileSize = 1, 
			edgeSize = 5, 
			insets   = {left = 0, right = 0, top = 0, bottom = 0},
		}
	);

	self.grid:SetBackdropColor( 0, 0, 0, 0.25 );

	if self.layout.show_grid == "no" then
		self.grid:Hide();
	end

	-- the button frame
	
	self.frame.class = self;

	self.frame:SetFrameLevel( 1 );
	self.frame:SetWidth( self.width );
	self.frame:SetHeight( self.height );
	self.frame:SetAllPoints( self.grid );
	
	-- scale and size the button

	self.btn_scale = self.height / self.button:GetHeight();
	
	-- relocate the button to our fake bar

	self.button.nUI_class = self;
	
	self.button:SetScale( self.btn_scale * (self.layout.rescale or 1));
	self.button:SetFrameStrata( "LOW" );
	self.button:SetFrameLevel( 1 );
	self.button:Show();

	self.button:ClearAllPoints();
	self.button:SetPoint( "BOTTOM", self.frame, "BOTTOM", 0, 0 );
	
	-- add this button to the nUI button cache
	
	local entry = 
	{ 
		frame  = self.frame, 
		obj    = self.button, 
		scale  = self.btn_scale, 
		width  = self.width, 
		height = self.height, 
	};

	nUI.Buttons[self.layout.name] = entry;
			
	-- cache references all of the relevant frames
	
	self.layers.flash     = _G[btn_name.."Flash"];
	self.layers.count     = _G[btn_name.."Count"];
	self.layers.name      = _G[btn_name.."Name"];
	self.layers.border    = _G[btn_name.."Border"];
	self.layers.cooldown  = _G[btn_name.."Cooldown"];
	self.layers.normal    = _G[btn_name.."NormalTexture"];
	self.layers.normal2   = _G[btn_name.."NormalTexture2"];
	self.layers.castable  = _G[btn_name.."AutoCastable"];
	self.layers.autocast  = _G[btn_name.."AutoCast"];
	self.layers.icon      = _G[btn_name.."Icon"];
	self.layers.hotkey    = _G[btn_name.."HotKey"];
	self.layers.pushed    = self.button:GetPushedTexture();
	self.layers.highlight = self.button:GetHighlightTexture();

	-- create a cooldown counter layer for it
	
	self.layers.cdcFrame:SetAllPoints( self.button );
	self.layers.cdcFrame:SetFrameLevel( 10 );
	
	self.layers.cdc:SetAllPoints( self.button );
	self.layers.cdc:SetJustifyH( "CENTER" );
	self.layers.cdc:SetJustifyV( "MIDDLE" );
	self.layers.cdc:SetFont( "Fonts/ARIALN.TTF", (self.height / self.btn_scale) * 0.52, "OUTLINE" );
	self.layers.cdc:SetTextColor( 1, 1, 0, 1 );
	
	-- tweak some of the layer scales to make them "fit" better
	
	if self.layers.border then
		local layer = self.layers.border;
		local inset = self.height * nUI.scale * 0.95;
		
		layer:ClearAllPoints();
		layer:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", -inset, inset );
		layer:SetPoint( "TOPRIGHT", self.frame, "TOPRIGHT", inset, inset );
		layer:SetPoint( "BOTTOMLEFT", self.frame, "BOTTOMLEFT", -inset, -inset );
		layer:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", inset, -inset );
	end

	if self.layers.autocast then
		
		self.layers.autocast:SetAlpha( 0.05 );
		
	end
		
	-- dispose of the default Bliz button outlines
	
	if not self.layout.safe then self.button:SetNormalTexture( "" ); end

	if self.layers.normal2 then self.layers.normal2:SetAlpha( 0 ); end
			
	if not self.custom_btn or not self.layers.hotkey then
					
		if self.layers.hotkey then 
			self.layers.hotkey:SetAlpha( 0 );
			self.layers.hotkey = self.layers.hotkey:GetParent():CreateFontString( self.layout.name.."_MyHotkey" );
		else
			self.layers.hotkeyFrame = CreateFrame( "Frame", self.layout.name.."_HotKeyFrame", self.button );
			self.layers.hotkey = self.layers.hotkeyFrame:CreateFontString( self.layout.name.."_MyHotKey" );				
		end
		
		self.layers.hotkey:ClearAllPoints();
		self.layers.hotkey:SetPoint( "TOP", self.button, "TOP", 1, 0 );
		self.layers.hotkey:SetPoint( "BOTTOM", self.button, "BOTTOM", 1, 1 );
		self.layers.hotkey:SetPoint( "LEFT", self.button, "LEFT", 1, 0 );
		self.layers.hotkey:SetPoint( "RIGHT", self.button, "RIGHT", 1, 0 );			
	end
	
	self.layers.hotkey:SetJustifyH( "LEFT" );
	self.layers.hotkey:SetJustifyV( "TOP" );
	self.layers.hotkey:SetFont( "Fonts/ARIALN.TTF", (self.height / self.btn_scale)/2.5, "OUTLINE" );
	self.layers.hotkey:SetTextColor( 1, 1, 1, 1 );

	if self.layers.count then
		self.layers.count:ClearAllPoints();
		self.layers.count:SetPoint( "TOP", self.button, "TOP", 0, 0 );
		self.layers.count:SetPoint( "BOTTOM", self.button, "BOTTOM", 0, 1 );
		self.layers.count:SetPoint( "LEFT", self.button, "LEFT", 1, 0 );
		self.layers.count:SetPoint( "RIGHT", self.button, "RIGHT", 1, 0 );
		self.layers.count:SetJustifyH( "RIGHT" );
		self.layers.count:SetJustifyV( "BOTTOM" );
		self.layers.count:SetFont( "Fonts/ARIALN.TTF", (self.height / self.btn_scale)/2.5, "OUTLINE" );
		self.layers.count:SetTextColor( 1, 1, 1, 1 );
	end

	if self.layers.name then
		self.layers.name:ClearAllPoints();
		self.layers.name:SetPoint( "TOP", self.button, "TOP", 1, 0 );
		self.layers.name:SetPoint( "BOTTOM", self.button, "BOTTOM", 1, 1 );
		self.layers.name:SetPoint( "LEFT", self.button, "LEFT", 1, 0 );
		self.layers.name:SetPoint( "RIGHT", self.button, "RIGHT", 1, 0 );
		self.layers.name:SetJustifyH( "CENTER" );
		self.layers.name:SetJustifyV( "BOTTOM" );
		self.layers.name:SetFont( "Fonts/ARIALN.TTF", (self.height / self.btn_scale)/3, "OUTLINE" );
		self.layers.name:SetTextColor( 1, 1, 1, 1 );
	end
	
	self:updateHotkey();
	self:onActionUsable();
	
	if self.button.action then
		local start, duration, enable = GetActionCooldown( self.button.action );
		self:startCooldown( start, duration, enable );
	end

	self:registerEvents();
end

-------------------------------------------------------------------------------

function prototype:registerEvents()
	
	self.frame:SetScript( "OnEvent", function() self:onEvent(); end );
	self.frame:RegisterEvent( "UPDATE_BINDINGS" );
	self.frame:RegisterEvent( "ACTIONBAR_SHOWGRID" );
	self.frame:RegisterEvent( "ACTIONBAR_HIDEGRID" );
	self.frame:RegisterEvent( "PLAYER_TARGET_CHANGED" );

end

-------------------------------------------------------------------------------

function prototype:unregisterEvents()
	
	self.frame:SetScript( "OnEvent", nil );
	self.frame:UnregisterEvent( "UPDATE_BINDINGS" );
	self.frame:UnregisterEvent( "ACTIONBAR_SHOWGRID" );
	self.frame:UnregisterEvent( "ACTIONBAR_HIDEGRID" );
	self.frame:UnregisterEvent( "PLAYER_TARGET_CHANGED" );

end

-------------------------------------------------------------------------------

function prototype:showGrid()

	self.grid:Show();
end

-------------------------------------------------------------------------------

function prototype:hideGrid()

	self.grid:Hide();
end

-------------------------------------------------------------------------------

function prototype:SetPoint( point, relative_to, relative_pt, xOfs, yOfs )
	
	self.grid:SetPoint( point, relative_to, relative_pt, xOfs, yOfs );
	
end

-------------------------------------------------------------------------------

function prototype:SetAllPoints( relative_to )
	
	self.grid:SetAllPoints( relative_to );
	
end

-------------------------------------------------------------------------------

function prototype:Show()
	
	self.button:Show();
	
end

-------------------------------------------------------------------------------

function prototype:Hide()
	
	self.button:Hide();
	
end

-------------------------------------------------------------------------------

function prototype:updateHotkey()
	
	if self.actionType then
	
		local button     = self.button;
		local actionName = "";
		
		local key1, key2 = GetBindingKey( self.actionType..self.actionID ) or
						   GetBindingKey( "CLICK "..button:GetName()..":LeftButton");
	
		local text = GetBindingText( key1 or key2, "KEY_", 1 );
		
		if ( text == "" ) then
			self.layers.hotkey:SetText( RANGE_INDICATOR );
			self.layers.hotkey:Hide();
		else
	
			text = text:gsub( "sc--9", "|c00FF9966NP|r(" );
			text = text:gsub( "sc--0", "|c00FF9966NP|r)" );
			text = text:gsub( "sc--", "|c00FF9966SC|r" );
			text = text:gsub( "c--", "|c00FF9966C|r" );
			text = text:gsub( "STRG--", "|c00CCCC00S|r" );
			text = text:gsub( "a--", "|c009966CCA|r" );
			text = text:gsub( "s--", "|c00CCCC00S|r" );
			text = text:gsub( "MAJ--", "|c00CCCC00S|r" );
			text = text:gsub( "Num Pad ", "|c00FF9966NP|r" );
			text = text:gsub( "Num Lock", "|c00FF9966NL|r" );
			text = text:gsub( "Mouse Button ", "M" );
			text = text:gsub( "Middle Mouse", "MM" );
			text = text:gsub( "Mouse Wheel Up", "WU" );
			text = text:gsub( "Mouse Wheel Down", "WD" );
			text = text:gsub( "Backspace", "Bs" );
			text = text:gsub( "Spacebar", "Sp" );
			text = text:gsub( "Delete", "De" );
			text = text:gsub( "Home", "Ho" );
			text = text:gsub( "End", "En" );
			text = text:gsub( "Insert", "Ins" );
			text = text:gsub( "Page Up", "Pu" );
			text = text:gsub( "Page Down", "Pd" );
			text = text:gsub( "Down Arrow", "D" );
			text = text:gsub( "Up Arrow", "U" );
			text = text:gsub( "Left Arrow", "L" );
			text = text:gsub( "Right Arrow", "R" );
		
			self.layers.hotkey:SetText( text );
			self.layers.hotkey:Show();
		end
	end
end

-------------------------------------------------------------------------------

function prototype:onEvent()
	
	if event == "UPDATE_BINDINGS" then
		self:updateHotkey();
	elseif event == "ACTIONBAR_SHOWGRID" then
		self:showGrid(); 
	elseif event == "ACTIONBAR_HIDEGRID" then
		self:hideGrid();
	elseif event == "PLAYER_TARGET_CHANGED" then
		self:onActionUsable();
	end

end

-------------------------------------------------------------------------------
-- called when ActionButton_OnUpdate is called on a button we have hooked

function prototype:onActionUpdate( elapsed )

	local do_action_update = false;
	local visible = true;

	-- check the button for visibility. If there's no icon for the button 
	-- or the icon is not shown, then we need to hide the cooldown count
	-- and hotkey. Likewise, there's no reason to do any other work if
	-- there's no visible buttons
	
	self.layers.icon = _G[self.layout.button.."Icon"];
	
	if not self.layers.icon then 
		visible = false;
	elseif not self.layers.icon:IsVisible() then
		visible = false;
	end

	if not visible then
	
		-- if we're not visible, then hide our custom layers, too
		
		if not self.hidden then
			self.layers.hotkey:SetAlpha( 0 );
			self.layers.cdc:SetAlpha( 0 );
			self.hidden = true;
		end
		
	else
		
		-- otherwise, if the layers are hidden, show them
		
		if self.hidden then
			
			self.layers.hotkey:SetAlpha( 1 );
			self.layers.cdc:SetAlpha( 1 );
			self.hidden = nil;
		end
	
		-- check for updates to the out of range and out of mana status
		
		self.Timers.range = self.Timers.range + elapsed;
		
		if self.Timers.range >= 0.08 then -- update at 12.5fsp
		
			self.Timers.range = 0;
			do_action_update  = true;
						
		end

		-- check for update to cooldown counter

		if self.Timers.cdc_expire then
			
			self.Timers.cdc = self.Timers.cdc + elapsed;
			
			if self.Timers.cdc >= self.Timers.cdc_expire then -- rate varies by time remaining
				
				self.Timers.cdc = 0;
			
				self:updateCooldown();
				
			end
		end
	end
	
	return do_action_update;

end

-------------------------------------------------------------------------------
-- called when ActionButton_UpdateUsable is called on a button we have hooked

function prototype:onActionUsable()
	
	-- there's nothing to do if there's no icon
	
	local update_vertex = false;
	local is_usable     = self.is_usable;
	local is_oor        = self.is_oor;
	local is_oom        = self.is_oom;
	local btn_name      = self.button:GetName();

	self.layers.icon    = _G[self.layout.button.."Icon"];

	if self.layers.icon and self.layers.icon:IsVisible() and self.layers.icon:GetAlpha() > 0 then
	
		local action       = self.button.action;

		if action then
			
			-- if the action has a range component, then we do oom, oor and unusable
			-- checks on it, otherwise, there's no point
			
			if ActionHasRange( action ) then
	
				local in_range     = IsActionInRange( action );			
				local can_use, oom = IsUsableAction( action );
				
				-- if the spell is usable then we need to clear the oom flag
				
				if can_use == 1 then
					
					is_oom = false;
					
				-- if the spell is not usable because of oom, be sure flag is set
						
				elseif oom == 1 then 
					
					is_oom = true;
			
				else
					
					is_usable = false;
					
				end
			
				if can_use == 1 or oom == 1 then
					
					-- if there's no target, then none of this means anything
					
					if not UnitExists( "target" ) then
						
						is_usable = true;
						is_oor    = false;
						in_range  = nil;
					
					-- if the spell is in range and we're flagged unusable or out of range, clear it
					
					elseif in_range == 1  then
						
						is_usable = true;
						is_oor    = false;
						
					-- if the spell is out of range and we're flagged unusable or not oor, fix it
					
					elseif in_range == 0 then
						
						is_usable = true;
						is_oor    = true;
								
					-- if the spell is unusable and we're not flagged as such, fix that
					
					else
						
						is_usable = false;
						
					end				
				end
			end
			
			-- if this is a consumable action and there's none left to consume, it's unusable
			
			if is_usable and IsConsumableAction( action ) then
				
				local count = GetActionCount( action );
				
				if count <= 0 then is_usable = false; end
				
			end
			
			-- if we changed any of the three states, save the state change
			
			if is_usable ~= self.is_usable or is_oor ~= self.is_oor or is_oom ~= self.is_oom then
				
				self.is_usable = is_usable;
				self.is_oor    = is_oor;
				self.is_oom    = is_oom;
	
				-- there is no built in mechanism I am aware of to check and see if a spell ID
				-- is usable or not... that is, you can test an action to see if it is usable
				-- or oom and you can check both actions and spells for range, but you cannot
				-- check a spell for usable or oom without an attached action. This hack creates
				-- a cache of spell states by spell ID as we update them on the action buttons.
				-- It isn't perfect, but it's usable. One place this is used is in range 
				-- checking since IsSpellInRange() can return 1 if the spell is unusable but
				-- the target is valid (i.e... hunter Kill Command will return "in range" any
				-- time the command is unusable even if the mob is out of range at the time.
				
				local _,id,_ = GetActionInfo( action );
				
				if id ~= nil then SpellStatus[id] =
				{
					usable = is_usable;
					oor    = is_oor;
					oom    = is_oom;
				}; end
			end				
	
			-- color the icon based on the state we last calculated for it
			-- I'd like to only do this when the state changes, but WoW's code
			-- steps on our coloring if we don't do it every time
			
			if not is_usable then 
				self.layers.icon:SetVertexColor( 0.35, 0.35, 0.35, 1 );
			elseif is_oor then
				local color = nUI_Config.Colors.OOR;			
				self.layers.icon:SetDesaturated( nil );
				self.layers.icon:SetVertexColor( color.r, color.g, color.b, 1 );
			elseif is_oom then
				local color = nUI_Config.Colors.OOM;			
				self.layers.icon:SetDesaturated( nil );
				self.layers.icon:SetVertexColor( color.r, color.g, color.b, 1 );
			else
				self.layers.icon:SetDesaturated( nil );
				self.layers.icon:SetVertexColor( 1, 1, 1, 1 );
			end
		end
	end	
end

-------------------------------------------------------------------------------

function prototype:startCooldown( start, duration, enable )

	-- start a new cooldown count
	
	if start > 0 and duration > 0 and enable > 0 then
		
		self.cdc_start    = start;
		self.cdc_duration = duration;
		
		self:updateCooldown();
		
	-- if we're not enabling and we have an active cooldown, cancel it
	
	elseif self.is_cdc then
		
		self.cdc_start    = GetTime();
		self.cdc_duration = -1;
		
		self:updateCooldown();
	
	end
end

-------------------------------------------------------------------------------
-- keep a cooldown counter on the face of applicable buttons

function prototype:updateCooldown()
	
	local button      = self.button;
	local action      = button.action;
	local type, id, _ = GetActionInfo( action );
	local cooldown    = Cooldowns[id];
	local cdc         = cooldown and cooldown.text or nil;
	local cdc2        = self.layers.cdc;
	local start       = self.cdc_start;
	local duration    = self.cdc_duration;
	local now         = GetTime();

	if not (start > 0) and not (duration > 0) and self.is_cdc then
			
		self.is_cdc            = false;
		self.is_flashing       = false;
		self.Timers.cdc_expire = nil;

		if cdc then cdc:SetText( "" ); end;
		
		cdc2:SetText( "" );
		
		if cooldown then 
			
			cooldown.done = true;
			if nUI.hud then nUI.hud:updateCooldowns(); end
		end
				
	-- we only process the cooldown after the current time has hit the start time
	
	elseif now < start then
		
		self.Timers.cdc_expire = start - now;
	
	-- otherwise we'll get to work
	
	else

		local remain     = duration - ( now - start );
		local text       = "";

		-- again... we only have work to do if there's still time remaining
		
		if remain <= 0 then
		
			-- if we have an active cooldown counter we need to expire it
			-- and we'll return to looking for cooldowns at 12.5fps
			
			if self.is_cdc then
			
				self.is_cdc            = false;
				self.is_flashing       = false;
				self.Timers.cdc_expire = nil;

				if cdc then cdc:SetText( "" ); end

				cdc2:SetText( "" );
				
				if cooldown then 
					cooldown.done = true;
					if nUI.hud then nUI.hud:updateCooldowns(); end
				end				
			end
			
		else

			if not self.is_cdc then
				
				local color = nUI_Config.Colors.CDC1;
				
				if type == "spell" then
					
					local name,_ = GetSpellName( id, BOOKTYPE_SPELL );
	
					-- if this is a spell that we know to be on the global cooldown counter
					-- then we can set the current value of the GCD accordingly. Note that
					-- is a value that can potentially change over time w/spellhaste and 
					-- the like.
					
					if nUI_GCDSpells[name] then nUI.GCD = duration; end
					
				end
				
				-- do not show counters on buttons that are just spinning the global cooldown
				
				if duration == (nUI.GCD or 1.5) then 
					if nUI.hud then nUI.hud:startGCD(); end
					return; 
				end

				-- this duration is not GCD, so run it...
				
				self.is_cdc      = true;
				self.is_flashing = false;

				if not cooldown then
					
					Cooldowns[id] = {};
					
					cooldown = Cooldowns[id];
				end
				
				cooldown.id       = id;
				cooldown.action   = action;
				cooldown.start    = start;
				cooldown.duration = duration;
				cooldown.done     = false;
				cooldown.icon     = self.layers.icon;

				cooldown.onEnter  = function() ActionButton_SetTooltip( self.button ); end; 
				cooldown.onLeave  = function() GameTooltip:Hide(); end;
				
				if nUI.hud then nUI.hud:updateCooldowns(); end
				
				cdc = cooldown.text;
				
				cdc2:SetTextColor( color.r, color.g, color.b, 1 );
				
			elseif not self.is_flashing and remain < 10 then
				
				local color      = nUI_Config.Colors.CDC2;
				self.is_flashing = true;
				
				if cdc then
					UIFrameFlash( cdc:GetParent(), 0.25, 0.25, remain-1, true, 0, 0.5 );
					cdc:SetTextColor( color.r, color.g, color.b, 1 );
				end

				UIFrameFlash( cdc2:GetParent(), 0.25, 0.25, remain-1, true, 0, 0.5 );
				cdc2:SetTextColor( color.r, color.g, color.b, 1 );

				-- flag the item as flashing in the cooldown table so the hud
				-- can modify the cooldown display
				
				if cooldown then 
					cooldown.flashing = true; 
					if nUI.hud then nUI.hud:updateCooldowns(); end
				end				
			end
			
			-- in the interest of efficiency, we set our time to expire for
			-- the next cooldown count check based on how much time is left.

			local text = nUI_DurationText( remain );
			
			if remain > 3600 then -- over an hour remaining
				
				self.Timers.cdc_expire = remain % 3600;
				
			elseif remain > 60 then -- over one minute remaining
			
				self.Timers.cdc_expire = remain % 60;
				
			elseif remain > 1 then -- more than a second to go
			
				self.Timers.cdc_expire = remain % 1;
				
			else -- less than a second
			
				self.Timers.cdc_expire = remain - math.floor( remain * 10 ) / 10;
				
			end
			
			-- if we have a text object, update it, otherwise we'll recheck at 12.5fps
			
			if cdc then cdc:SetText( text ); cdc2:SetText( "" );
			elseif not cooldown then self.Timers.cdc_expire = 0.08;
			else cdc2:SetText( text );
			end
		end
	end
end
